use crate::constants;
use std::os::linux::fs::MetadataExt;
use std::path::Path;
use std::process::Command;
use std::time::UNIX_EPOCH;

#[derive(thiserror::Error, Debug)]
enum DisksError {
    #[error("DiskParseError: {0}")]
    Parse(&'static str),
}

#[derive(Debug, serde::Serialize)]
#[serde(tag = "format")]
pub enum DiskFileFormat {
    Raw { is_sparse: bool },
    QCow2 { virtual_size: usize },
    CompressedRaw,
    CompressedQCow2,
}

/// Disk file information
#[derive(serde::Serialize)]
pub struct DiskFileInfo {
    pub file_size: usize,
    #[serde(flatten)]
    pub format: DiskFileFormat,
    pub file_name: String,
    pub name: String,
    pub created: u64,
}

impl DiskFileInfo {
    /// Get disk image file information
    pub fn load_file(file: &Path) -> anyhow::Result<Self> {
        if !file.is_file() {
            return Err(DisksError::Parse("Path is not a file!").into());
        }

        // Get file metadata
        let metadata = file.metadata()?;
        let mut name = file
            .file_stem()
            .and_then(|s| s.to_str())
            .unwrap_or("disk")
            .to_string();
        let ext = file.extension().and_then(|s| s.to_str()).unwrap_or("raw");

        // Determine file format
        let format = match ext {
            "qcow2" => DiskFileFormat::QCow2 {
                virtual_size: qcow_virt_size(file)?,
            },
            "raw" => DiskFileFormat::Raw {
                is_sparse: metadata.len() / 512 >= metadata.st_blocks(),
            },
            "gz" if name.ends_with(".qcow2") => {
                name = name.strip_suffix(".qcow2").unwrap_or(&name).to_string();
                DiskFileFormat::CompressedQCow2
            }
            "gz" => DiskFileFormat::CompressedRaw,
            _ => anyhow::bail!("Unsupported disk extension: {ext}!"),
        };

        Ok(Self {
            name,
            file_size: metadata.len() as usize,
            format,
            file_name: file
                .file_name()
                .and_then(|s| s.to_str())
                .unwrap_or("")
                .to_string(),
            created: metadata
                .created()?
                .duration_since(UNIX_EPOCH)
                .unwrap()
                .as_secs(),
        })
    }
}

#[derive(serde::Deserialize)]
struct QCowInfoOutput {
    #[serde(rename = "virtual-size")]
    virtual_size: usize,
}

/// Get QCow2 virtual size
fn qcow_virt_size(path: &Path) -> anyhow::Result<usize> {
    // Run qemu-img
    let mut cmd = Command::new(constants::QEMU_IMAGE_PROGRAM);
    cmd.args([
        "info",
        path.to_str().unwrap_or(""),
        "--output",
        "json",
        "--force-share",
    ]);
    let output = cmd.output()?;
    if !output.status.success() {
        anyhow::bail!(
            "{} info failed, status: {}, stderr: {}",
            constants::QEMU_IMAGE_PROGRAM,
            output.status,
            String::from_utf8_lossy(&output.stderr)
        );
    }
    let res_json = String::from_utf8(output.stdout)?;

    // Decode JSON
    let decoded: QCowInfoOutput = serde_json::from_str(&res_json)?;
    Ok(decoded.virtual_size)
}
